/** @file sample_apci3120.c
*
* @author Sylvain Nahas
*
* This software 
* - scans the /proc/sys/apci3120/ directory 
* - run a serie of test on each detected boards
*/

/** @par LICENCE
* @verbatim
 Copyright (C) 2004,2005  ADDI-DATA GmbH for the source code and the documentation.
        
        ADDI-DATA GmbH
        Dieselstrasse 3 
        D-77833 Ottersweier
        Tel: +19(0)7223/9493-0
        Fax: +49(0)7223/9493-92
        http://www.addi-data-com
        info@addi-data.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. 
        
    You shoud find the complete GPL in the COPYING file accompanying 
    this source code. 
* @endverbatim
*/


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <sys/sysmacros.h> /* major() minor() */

#include <apci3120.h>

//------------------------------------------------------------------------------
/** Filter the "." and ".." entries in the struct dirent ** area set up by the scandir() function (utility function).*/
int
get_real_dir_entry (const struct dirent * ent)
{
        if (! strcmp(ent->d_name,"."))
                return 0;
        if (! strcmp(ent->d_name,".."))
                return 0;
                                
        return 1;
}

//--------------------------------------------------------------------------------
/** this function look up for cards in the /proc/sys/apci3120/ directory 
* @param cards A pointer to the card area to fill
* @return The number of cards detected (may be zero if nor cards found).
*/
int apci3120_find_cards(int ** cards)
{
	
	struct dirent **eps;
	int cnt = 0; // number of card found
	char APCI_SPECIAL_FILE_DIR[255]={0}; 
	char proc_path_template[255]={0};

	/* Test which node is availabe */
	{
		struct stat file_stat;

		if (stat("/proc/sys/apci3120",&file_stat) == 0)
		{
			/* Kernel 2.4 */
			strncpy(APCI_SPECIAL_FILE_DIR, "/proc/sys/apci3120/", 19);
			strncpy(proc_path_template, "%s%s/ctrl", 9);
		}
		else if (stat("/dev/apci3120",&file_stat) == 0)
		{
			/* Kernel 2.6 */
			strncpy(APCI_SPECIAL_FILE_DIR, "/dev/apci3120/", 14);
			strncpy(proc_path_template, "%s%s", 4);
		}
		else
		{
			perror("fstat");
			exit(1);
		}
	}
	
	/* scan directory for card entry */
	if ( (cnt = scandir (APCI_SPECIAL_FILE_DIR, &eps, get_real_dir_entry , alphasort)) == -1 )
	{
		perror("scandir");
		return 0;
	}	
	
	if (cnt == 0) return 0;
	
	/* allocate cards area */
	if ( ! ( (*cards) = malloc( cnt * sizeof(int) ) ) )
	{
		perror("malloc");
		free(eps);
		return 0;
	}
	
	/* for each card entry, open it and store file descriptor in cards[] */
	{
		int i;
		for (i=0; i< cnt; i++)
		{
			int fd;
			char name[100];
			
			snprintf(name,sizeof(name),proc_path_template,APCI_SPECIAL_FILE_DIR,eps[i]->d_name);
			
			if ( ( fd = open(name, O_RDWR)) == -1 )
			{
				perror("open");
				exit(1);
			}			
									
			{
				struct stat file_stat;
				if ( fstat(fd,&file_stat) )
				{
					perror("fstat");
					exit(1);
				}			
				if (!S_ISCHR(file_stat.st_mode)) 
				{
					/* should never happen !*/
					printf("file %d is not a character device\n",i);
					exit(1);
				}
				printf("%s, major:%d minor:%d \n",name, major(file_stat.st_rdev),minor(file_stat.st_rdev));
			
				(*cards)[minor(file_stat.st_rdev)] = fd;
			} 
			/* index device file descriptor by the minor number */			
		}	
	}
		
	return cnt;
}
//--------------------------------------------------------------------------------

void apci3120_run_test(int fd)
{
	/* Test analog inputs */
	printf("=> CMD_APCI3120_Read1AnalogInput\n");
	{
		unsigned int arg[5] = {0};
		int error = 0;
		unsigned int channel = 0;
		
		arg[0] = channel;		/* Channel number (1 to 16) 								*/
		arg[1] = APCI3120_1_GAIN;	/* Gain to use (APCI3120_1_GAIN, APCI3120_2_GAIN, APCI3120_5_GAIN and APCI3120_10_GAIN) */
		arg[2] = APCI3120_BIPOLAR;	/* Polarity can be APCI3120_UNIPOLAR or APCI3120_BIPOLAR 				*/
		arg[3] = 10;			/* Selection of the conversion time from 10�s to 32767�s 				*/	
		arg[4] = APCI3120_DISABLE;	/* Interrupt flag (NOT IMPLEMENTED)							*/

		error = ioctl (fd, CMD_APCI3120_Read1AnalogInput, arg);

		switch (error)
		{			
			case 0: printf ("Value for channel %i = %u.\n", channel, arg[0]);
			break;

			case 1: printf ("The handle parameter of the board is wrong.\n");
			break;

			case 2: printf ("Number of the analog input is wrong.\n");
			break;

			case 3: printf ("The selected gain is wrong.\n");
			break;

			case 4: printf ("The selected input voltage range is wrong.\n");
			break;

			case 5: printf ("The selected conversion time is wrong.\n");
			break;

			case 6: printf ("A wrong parameter has been passed for b_InterruptFlag.\n");
			break;

			case 7: printf ("Time out.\n");
			break;

			default:perror("ioctl");
			break;
		}
	}

	/* Test analog inputs */
	printf("=> CMD_APCI3120_ReadMoreAnalogInputs\n");
	{
		unsigned int arg[(16*3) + 1 + 1] = {0};
		int error = 0;
		int i = 0;

		/* Number of channel to read (1 - 16) */
		arg[0] = 16;

		/* List of channels to read, here channel 0 to 15 */
		for (i=0; i<arg[0]; i++)
			arg[(1+i)] = i;

		/* Set the gain for each channel */
		for (i=0; i<arg[0]; i++)
			arg[(arg[0] + 1 + i)] = APCI3120_1_GAIN;

		/* Set the polarity for each channel */
		for (i=0; i<arg[0]; i++)
			arg[(arg[0] + arg[0] + 1 + i)] = APCI3120_BIPOLAR;

		/* Set the converting time (it is the same for each channel) it can be of 10 up to 32767 us*/
		arg[(arg[0] + arg[0] + arg[0] + 1)] = 10;

		error = ioctl (fd, CMD_APCI3120_ReadMoreAnalogInputs, arg);

		switch (error)
		{			
			case 0: 
				for (i=0; i<16; i++)
					printf ("Channel %i = %u\n", i, arg[i]);
			break;

			case 1: printf ("The handle parameter of the board is wrong.\n");
			break;

			case 2: printf ("The size of the sequence table is wrong.\n");
			break;

			case 3: printf ("Wrong parameter found in table the channel array\n");
			break;

			case 4: printf ("Wrong parameter found in table the gain array\n");
			break;

			case 5: printf ("Wrong parameter found in table the polarity array\n");
			break;

			case 6: printf ("he selected conversion time is wrong.\n");
			break;

			case 7: /* Can not occured */
			break;

			case 8: printf ("Time out.\n");
			break;

			default:perror("ioctl");
			break;
		}
	}

	/* Test analog outputs */
	printf("=> CMD_APCI3120_Write1AnalogOutput\n");
	{
		unsigned int arg[3] = {0};
		int error = 0;
		unsigned int channel = 0;
		
		arg[0] = channel;		/* Channel number (1 to 16) 								*/
		arg[1] = APCI3120_BIPOLAR;	/* Polarity can be APCI3120_UNIPOLAR or APCI3120_BIPOLAR 				*/
		arg[2] = 16383;			/* Value to write (Unipolar 0 - 8192, Bipolar 0 - 16383) 				*/
	
		error = ioctl (fd, CMD_APCI3120_Write1AnalogOutput, arg);
	
		switch (error)
		{			
			case 0: printf ("Write %u on channel %i.\n", arg[2], channel);
			break;

			case 1: printf ("The handle parameter of the board is wrong.\n");
			break;

			case 2: printf ("Number of the analog output is wrong.\n");
			break;

			case 3: printf ("The selected output voltage range is wrong.\n");
			break;

			case 4: printf ("Wrong output value for the output channel.\n");
			break;

			case 5: printf ("Time out.\n");
			break;

			default:perror("ioctl");
			break;
		}
	}	

	/* Test analog inputs */
	printf("=> CMD_APCI3120_WriteMoreAnalogOutput\n");
	{
		unsigned int arg[(8*2) + 1 + 1] = {0};
		int error = 0;
		int i = 0;

		/* First channel to write */
		arg[0] = 0;

		/* Number of channels to write */
		arg[1] = 8;

		/* Set the polarity for each channel */
		for (i=0; i<arg[1]; i++)
			arg[(2 + i)] = APCI3120_BIPOLAR;

		/* Set the value to write on each channel */
		for (i=0; i<arg[1]; i++)
			arg[(2 + i + 8)] = 16383;

		error = ioctl (fd, CMD_APCI3120_WriteMoreAnalogOutput, arg);

		switch (error)
		{			
			case 0: 
				for (i=0; i<arg[1]; i++)
					printf ("Write %u on channel %i\n", arg[(2 + i + 8)], (i + arg[0]));
			break;

			case 1: printf ("The handle parameter of the board is wrong.\n");
			break;

			case 2: printf ("Number of the analog output is wrong.\n");
			break;

			case 3: printf ("The selected output voltage range is wrong.\n");
			break;

			case 4: printf ("Wrong output value for the output channel.\n");
			break;

			case 5: printf ("Time out.\n");
			break;

			default:perror("ioctl");
			break;
		}
	}
}
//--------------------------------------------------------------------------------
int main(int argc, char** argv)
{

	int * apci3120_card_fd;
	unsigned int apci3120_card_number = 0;
	{
		apci3120_card_number = apci3120_find_cards(&apci3120_card_fd);
		
		printf("total: %d cards\n",apci3120_card_number);
	}	
	
	{
		int i;
		for (i=0;i< apci3120_card_number; i++)
		{
			printf("\n+++++++++++++++++++++++++++++++++++++\nrunning test on device of minor number %d\n+++++++++++++++++++++++++++++++++++++\n\n\n",i);
			apci3120_run_test(apci3120_card_fd[i]);	
		}
	}

	

	return 0;
}
//--------------------------------------------------------------------------------
